#include "MoveGeneration.h"
#include "ChessGame.h"

using namespace chessy;

template class PawnPushGenerator<WHITE_SIDE>;
template class PawnPushGenerator<BLACK_SIDE>;


std::string Move::asString(void) const
{
    return origin.getName() + "-" + destination.getName();
}


std::vector<Move> MoveGenerator::generatePossibleMoves(void) const
{
    const ChessGame& cg = game;
    Bitboard whites = cg.getBoard().getWhites();
    Bitboard blacks = cg.getBoard().getBlacks();
    std::vector<Move> ret;
    PawnPushGenerator<BLACK_SIDE> mg{ cg };
    for (auto push : mg)
        ret.push_back(push);
    PositionSet knights = cg.getBoard().getBlackKnights();
    for (auto knight : knights) {
        for (auto pos : KnightMoveGenerator{ knight, blacks }) {
            ret.push_back(Move{ knight, pos });
        }
    }

    PositionSet bishops = cg.getBoard().getBlackBishops();
    for (auto bishop : bishops) {
        for (auto pos : PrimitiveBishopMoveGenerator{ bishop, whites, blacks }) {
            ret.push_back(Move{ bishop, pos });
        }
    }

    PositionSet rooks = cg.getBoard().getBlackRooks();
    for (auto rook : rooks) {
        for (auto pos : PrimitiveRookMoveGenerator{ rook, whites, blacks }) {
            //cout << "rook: " << Move(rook, pos).asString() << endl;
        }
    }

    PositionSet queens = cg.getBoard().getBlackQueens();
    for (auto queen : queens) {
        for (auto pos : PrimitiveQueenMoveGenerator{ queen, whites, blacks }) {
            //cout << "queen: " << Move(queen, pos).asString() << endl;
        }
    }

    Bitboard king = cg.getBoard().getBlackKings();
    Index kingIndex = king.getLeastSignificantBit();
    for (auto pos : KingMoveGenerator{ king, blacks }) {
        //cout << "king: " << Move(kingIndex, pos).asString() << endl;
    }
    return ret;
}


template<int side>
typename PawnPushGenerator<side>::MoveIterator PawnPushGenerator<side>::begin(void) const
{
    const Board& board = chessGame.getBoard();
    Bitboard movedPieces = side == WHITE_SIDE ?
        board.getWhitePawns() :
        board.getBlackPawns();
    if (side == WHITE_SIDE)
        movedPieces.moveNorthOne();
    else
        movedPieces.moveSouthOne();
    
    movedPieces &= ~(side == WHITE_SIDE ? board.getBlacks() : board.getWhites());

    return MoveIterator{ movedPieces };
}


template<int side>
typename PawnPushGenerator<side>::MoveIterator PawnPushGenerator<side>::end(void) const
{
    return MoveIterator{ 0 };
}


const std::array<Bitboard, 64> KnightMoveGenerator::moveSets = generateKnightMoves();

Bitboard KnightMoveGenerator::generateFromIndex(Index i) {
    int row = i.getRow();
    int column = i.getColumn();
    Bitboard result = 0;
    auto makeBitboardIfValid = [](int row, int column) -> Bitboard {
        return row < 8 && column < 8 && row >= 0 && column >= 0 ?
            Bitboard::fromIndex({row, column}) : Bitboard{ 0 };
    };
    result |= makeBitboardIfValid(row - 2, column - 1);
    result |= makeBitboardIfValid(row - 1, column - 2);
    result |= makeBitboardIfValid(row + 2, column - 1);
    result |= makeBitboardIfValid(row + 1, column - 2);
    result |= makeBitboardIfValid(row - 2, column + 1);
    result |= makeBitboardIfValid(row - 1, column + 2);
    result |= makeBitboardIfValid(row + 2, column + 1);
    result |= makeBitboardIfValid(row + 1, column + 2);
    return result;
}


std::array<Bitboard, 64> KnightMoveGenerator::generateKnightMoves(void)
{
    std::array<Bitboard, 64> moves{};
    for (int8_t i = 0; i < moves.size(); i++) {
        moves[i] = generateFromIndex(Index{ i });
    }
    return moves;
}


Bitboard PrimitiveQueenMoveGenerator::generateAttackSet(Index i,
    Bitboard enemies, Bitboard friendly, bool straight, bool diagonal)
{
    Bitboard queen = Bitboard::fromIndex(i);
    Bitboard occupied = (enemies | friendly) & ~queen;

    // strange c++ incoming
    auto iterate = [queen, occupied] (void (Bitboard::*moveStep)()) -> Bitboard {
        Bitboard result = 0;
        int a = 0;
        for (Bitboard r = queen; !(r & occupied) && r;) {
            (r.*moveStep)();
            result |= r;
            a++;
        }
        return result;
    };

    Bitboard possibleMoves = 0;
    if (straight)
        possibleMoves |=
            iterate(&Bitboard::moveNorthOne) |
            iterate(&Bitboard::moveSouthOne) |
            iterate(&Bitboard::moveEastOne) |
            iterate(&Bitboard::moveWestOne);
    if (diagonal)
        possibleMoves |=
            iterate(&Bitboard::moveNEOne) |
            iterate(&Bitboard::moveSEOne) |
            iterate(&Bitboard::moveNWOne) |
            iterate(&Bitboard::moveSWOne);   
    return (possibleMoves & ~friendly);
}


Bitboard KingMoveGenerator::generateKingMoves(Bitboard king)
{
    return king.northOne() | king.southOne() | king.eastOne() | king.westOne() |
        king.neOne() | king.nwOne() | king.swOne() | king.seOne();
}

