#ifndef CHESSY_MOVEGENERATION_H
#define CHESSY_MOVEGENERATION_H

#include "BitBoard.h"
#include <string>
#include <vector>
#include <array>

namespace chessy
{
    // forward declaration
    class ChessGame;

    struct Move;

    class MoveGenerator;

    class PositionSet;
    template<int side>
    class PawnPushGenerator;
    class KnightMoveGenerator;
    class PrimitiveQueenMoveGenerator;
    class PrimitiveRookMoveGenerator;
    class PrimitiveBishopMoveGenerator;
    class KingMoveGenerator;

    const int WHITE_SIDE = 0;
    const int BLACK_SIDE = 1;
}


struct chessy::Move
{
    Index origin;
    Index destination;

    Move            (void)          = default;
    Move            (const Move&)   = default;
    Move            (Move&&)        = default;
    ~Move           (void)          = default;
    Move& operator= (const Move&)   = default;
    Move& operator= (Move&&)        = default;

    inline Move(Index origin, Index destination) :
        origin{ origin }, destination{ destination } {}

    std::string asString(void) const;
};


class chessy::MoveGenerator
{
    const ChessGame& game;
public:
    inline MoveGenerator(const ChessGame& game) :
        game{ game } {}

    std::vector<Move> generatePossibleMoves(void) const;
};


class chessy::PositionSet
{
    Bitboard bitboard;
public:

    inline PositionSet(Bitboard b) : bitboard{ b } {}

    struct PositionSetIterator
    {
        Bitboard bitboard;

        inline Index operator *(void) const
        {
            return Index{ int8_t(trailingZeroes(bitboard.bits)) };
        }

        inline void operator ++(void)
        {            
            bitboard.bits &= bitboard.bits - 1ULL; // remove least significant one bit
        }

        inline bool operator !=(const PositionSetIterator& psi) const
        {
            return bitboard != psi.bitboard;
        }
    };

    inline PositionSetIterator begin(void) const
    {
        return PositionSetIterator{ bitboard };
    }

    inline static PositionSetIterator end(void)
    {
        return PositionSetIterator{ 0 };
    }
};


template<int side>
class chessy::PawnPushGenerator
{
    const ChessGame& chessGame;

    struct MoveIterator
    {
        PositionSet::PositionSetIterator pawnPushes;
        inline Move operator *(void) const
        {
            Index pp = *pawnPushes;
            return Move{ int8_t(pp + (side != WHITE_SIDE ? 8 : -8)), pp };
        }

        inline void operator ++(void)
        {
            ++pawnPushes;
        }

        inline bool operator !=(const MoveIterator& psi) const
        {
            return pawnPushes != psi.pawnPushes;
        }
    };
public:
    inline PawnPushGenerator(const ChessGame& cg) : chessGame{ cg } {}

    MoveIterator begin(void) const;
    MoveIterator end(void) const;
};


/*!
 * extends \link PositionSet so that all possible destinations for
 * a knight can be iterated over.
 */
class chessy::KnightMoveGenerator :
    public PositionSet
{
    static const std::array<Bitboard, 64> moveSets;

public:
    inline KnightMoveGenerator(Index knight, Bitboard occupied) :
        PositionSet{ moveSets[knight] & ~occupied } {}

private:
    static Bitboard generateFromIndex(Index i);
    static std::array<Bitboard, 64> generateKnightMoves(void);
};


/*!
 * uses a primitive algorithm to calculate all possible move destinations
 * for a queen.
 */
class chessy::PrimitiveQueenMoveGenerator :
    public PositionSet
{
protected:
    inline PrimitiveQueenMoveGenerator(Index i, Bitboard enemies, Bitboard friendly,
        bool straight, bool diagonal) :
        PositionSet{ generateAttackSet(i, enemies, friendly, straight, diagonal) } {}
public:
    inline PrimitiveQueenMoveGenerator(Index i, Bitboard enemies,
        Bitboard friendly) :
        PrimitiveQueenMoveGenerator{ i, enemies, friendly, true, true } {}
private:
    Bitboard generateAttackSet(Index i, Bitboard enemies, Bitboard friendly,
        bool straight, bool diagonal);
};


class chessy::PrimitiveBishopMoveGenerator :
    public PrimitiveQueenMoveGenerator
{
public:
    inline PrimitiveBishopMoveGenerator(Index i, Bitboard enemies,
        Bitboard friendly) :
        // bishop moves are actually just diagonal queen moves
        PrimitiveQueenMoveGenerator{ i, enemies, friendly, false, true } {}
};


class chessy::PrimitiveRookMoveGenerator :
    public PrimitiveQueenMoveGenerator
{
public:
    inline PrimitiveRookMoveGenerator(Index i, Bitboard enemies,
        Bitboard friendly) :
        // rook moves are actually just straight queen moves
        PrimitiveQueenMoveGenerator{ i, enemies, friendly, true, false } {}
};


class chessy::KingMoveGenerator :
    public PositionSet
{
public:
    inline KingMoveGenerator(Bitboard king, Bitboard friendly) :
        PositionSet{ generateKingMoves(king) & ~friendly } {}
private:
    Bitboard generateKingMoves(Bitboard king);
};

#endif // CHESSY_MOVEGENERATION_H
